import json
import math
import time
import threading
import queue
from tweepipe.legacy.utils import extract

import tweepy
from loguru import logger

from tweepipe.db import db_client
from tweepipe.utils import tracker, credentials


def stream_keywords(
    twitter_credentials: dict,
    keywords: list = [],
    issue: str = "stream",
    include_relations: bool = True,
    include_users: bool = True,
    output_file: str = None,
    skip_db: bool = False,
):

    stream_log_file = f"seq_stream_log_{issue}.log"
    logger.add(stream_log_file)
    logger.info(f"Starting streaming for {issue}, saving at {stream_log_file}.")

    try:
        if not skip_db:
            db_conn = db_client.DBClient(
                issue=issue,
                include_relations=include_relations,
                include_users=include_users,
            )
        else:
            db_conn = None
    except Exception:
        db_conn = None
        pass

    # implement exponential backoff in case of trouble
    exp_backoff, backoff_param = 1, 1.4
    while True:
        try:
            # reset the exponential backoff when going beyond 15 minutes
            if exp_backoff > 900:
                exp_backoff = 1.4

            print("credentials", twitter_credentials)
            twitter_auth = credentials._get_twitter_auth(twitter_credentials)
            stream_listener = SNPipelineStream(
                db_conn=db_conn,
                output_file=output_file,
                include_users=include_users,
                include_relations=include_relations,
            )
            stream = tweepy.Stream(auth=twitter_auth, listener=stream_listener)
            stream.filter(track=keywords, stall_warnings=True)
        except Exception as e:
            error_msg = f"Error while streaming with {len(keywords)} keywords., waiting {exp_backoff} seconds..."
            tracker.log_error(e, msg=error_msg)
            logger.error(error_msg)
            time.sleep(exp_backoff)

            # backing off with an exponential factor of 1.4
            exp_backoff = exp_backoff * backoff_param


def _run_streamer(
    twitter_auth: dict,
    keywords: list = [],
    issue: str = "stream",
    include_relations: bool = True,
    include_users: bool = True,
):
    """Helper function dedicated to running streaming processes in parallel."""
    stream_log_file = f"stream_log_{issue}.log"
    logger.add(stream_log_file)
    logger.info(f"Starting streaming for {issue}, saving at {stream_log_file}.")

    db_conn = db_client.DBClient(
        issue=issue, include_relations=include_relations, include_users=include_users
    )
    stream_listener = SNPipelineStream(db_conn=db_conn)
    stream = tweepy.Stream(auth=twitter_auth, listener=stream_listener)
    stream.filter(track=keywords, stall_warnings=True)


def _get_parallel_streaming_kwargs(
    keywords: list,
    twitter_auths: list,
    process_count: int,
    issue: str,
    include_users: bool = True,
    include_relations: bool = True,
):
    """Retrieve a list of keyword args for streaming keywords in parallel. These
    include twitter_auth accesses, the list of keywords to stream as well as
    the number of processes to use and type of information to retrieve.
    """

    batch_size = math.ceil(len(keywords) / process_count)
    keywords_batches = [
        keywords[i : i + batch_size] for i in range(0, len(keywords), batch_size)
    ]

    kwargs_list = []
    for keyword_batch, twitter_auth in zip(keywords_batches, twitter_auths):
        tmp_kwargs = dict(
            twitter_auth=twitter_auth,
            keywords=keyword_batch,
            issue=issue,
            include_users=include_users,
            include_relations=include_relations,
        )
        kwargs_list.append(tmp_kwargs)

    return kwargs_list


class SNPipelineStream(tweepy.StreamListener):
    def __init__(
        self,
        db_conn: db_client.DBClient = None,
        output_file: str = None,
        include_users: bool = True,
        include_relations: bool = True,
    ):
        super(SNPipelineStream, self).__init__()

        self.db_conn = db_conn
        self.output_file = output_file
        self.include_users = include_users
        self.include_relations = include_relations
        self.processing_queue = queue.Queue()
        self.processing_thread = threading.Thread(
            target=self.process_incoming_tweets, daemon=True
        )
        self.processing_thread.start()

    def process_incoming_tweets(self, heartbeat: float = 1):
        while True:
            if self.processing_queue.empty():
                # logger.debug(f"No tweet, waiting {heartbeat} sec.")
                time.sleep(heartbeat)
                continue
            else:
                try:
                    tweet = self.processing_queue.get(block=False)
                    # logger.debug(
                    #    f'Processing tweet {tweet.get("id")} - queue size {self.processing_queue.qsize()} -batch size {len(self.db_conn.tweet_batch)}.'
                    # )
                    extract.retrieve_content_from_tweet(
                        tweet=tweet,
                        db_conn=self.db_conn,
                        include_users=self.include_users,
                        include_relations=self.include_relations,
                        output_file=self.output_file,
                    )
                    self.processing_queue.task_done()
                except queue.Empty as e:
                    logger.error(e)
                    continue

    def on_status(self, status):
        """Add loaded tweet status to the queue of incoming tweet to be
        processed."""
        try:
            # logger.debug(f"Adding tweet {status.id_str} to processing queue.")
            self.processing_queue.put(status._json)
        except queue.Full as e:
            logger.error(f"Error: {e}, queue size: {self.processing_queue.qsize()}.")
            raise e

        return True

    def keep_alive(self):
        """Called when a keep-alive arrived"""
        return

    def dump_processing_queue_content(self):
        tweets = []
        while not self.processing_queue.empty():
            try:
                tweets.append(self.processing_queue.get_nowait())
            except queue.Empty as e:
                logger.error(e)

        with open("emergency_tweet_queue_dump.json", "w") as f:
            json.dump(tweets, f)

    def on_exception(self, exception):
        """Called when an unhandled exception occurs."""
        # TODO: check if should perform following operation
        try:
            self.db_conn.flush_content()
        except Exception as e:
            pass
        self.dump_processing_queue_content()
        logger.error(exception)
        raise Exception(exception)

    # TODO: implement sleeping so avoid getting
    def on_limit(self, track):
        """Called when a limitation notice arrives"""
        try:
            self.db_conn.flush_content()
        except Exception as e:
            pass
        error_msg = f"Reached credential limit {track}."
        logger.warning(error_msg)
        raise TimeoutError(error_msg)

    def on_error(self, status_code):
        """Called when a non-200 status code is returned"""
        try:
            self.db_conn.flush_content()
        except Exception as e:
            pass
        error_msg = f"Received error on Twitter stream with code: {status_code}."
        logger.error(error_msg)
        raise Exception(error_msg)

    def on_timeout(self):
        """Called when stream connection times out"""
        try:
            self.db.flush_content()
        except Exception as e:
            pass
        timeout_msg = "Twitter connection timed out."
        logger.error(timeout_msg)
        raise TimeoutError(timeout_msg)

    def on_disconnect(self, notice):
        """Called when twitter sends a disconnect notice
        Disconnect codes are listed here:
        https://developer.twitter.com/en/docs/tweets/filter-realtime/guides/streaming-message-types
        """
        try:
            self.db_conn.flush_content()
        except Exception as e:
            pass
        disconnect_msg = f"Twitter client disconnecting with notice: {notice}."
        logger.error(disconnect_msg)
        raise Exception(disconnect_msg)

    def on_warning(self, notice):
        """Called when a disconnection warning message arrives"""
        self.db_conn.flush_content()
        warning_msg = f"Warning on Twitter Stream: {notice}."
        logger.warning(warning_msg)


# future snpipeline stream module when tweepy 4.0 releases
class SNPipelineStreamFuture(tweepy.Stream):
    def __init__(
        self,
        credentials: dict,
        db_conn: db_client.DBClient = None,
        output_file: str = None,
    ):
        # TODO: check if super init valid
        super(SNPipelineStreamFuture, self).__init__(**credentials)
        self.db_conn = db_conn
        self.output_file = output_file

    def on_data(self, data: str):
        tweet = json.loads(data)

        if "id_str" not in tweet:
            return True

        extract.retrieve_content_from_tweet(
            tweet,
            db_conn=self.db_conn,
            output_file=self.output_file,
            include_users=self.db_conn.include_users,
            include_relations=self.db_conn.include_relations,
        )

        return True



import os

from setuptools import find_packages, setup

import tweepipe


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="tweepipe",
    description=".",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=tweepipe.__version__,
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[],
    license="MIT",
)



import json
import pickle
import math

import jsonlines
from loguru import logger

from tweepipe import settings
from tweepipe.db import db_client
from tweepipe.utils import errors


def _save_to_json(content=None, output_file=None, mode="w"):
    """Simple wrapper to write content to a given json file."""

    with open(output_file, mode) as f:
        json.dump(content, f)


def _save_to_jsonl(content: list = [], output_file: str = None, mode: str = "w"):
    """Wrapper to write lists to a jsonl file (successive lines)."""

    with open(output_file, mode) as f:
        with jsonlines.Writer(f) as writer:
            writer.write_all(content)


def _save_to_csv(content: list, output_file: str):
    with open(output_file, "w") as f:
        for line in content:
            f.write(f"{line}\n")


def _get_loading_tids_kwargs(files, env_file, batch_size, issue, worker_count):
    kwargs_list = []
    filebatch_size = math.ceil(len(files) / worker_count)
    for i in range(0, len(files), filebatch_size):
        kwargs = {
            "files": files[i : i + filebatch_size],
            "env_file": env_file,
            "batch_size": batch_size,
            "issue": issue,
        }
        kwargs_list.append(kwargs)

    return kwargs_list


def load_uids_from_file_to_db(
    file: str,
    db_conn: db_client.DBClient,
    batch_size: int = 2048,
    env_file: str = None,
    api_count: int = 1,
):
    """
    Load uids from file to db for hydration.

    Args:
        file (str): Uid filename.
        db_conn (db_client.DBClient): Active db connection.
        batch_size (int, optional): Batch size for db updates. Defaults to 2048.
        env_file (str, optional): Dotenv file. Defaults to None.
        api_count (int, optional): Number of APIs to employ when hydrating. Defaults to 1.
    """
    settings.load_config(env_file)
    uids = _load_from_file(file)
    keys = list(range(1, api_count + 1))

    for i in range(0, len(uids), batch_size):
        uid_docs_batch = [
            {"uid": uid, "status": 0, "key": keys[i % api_count]}
            for uid in uids[i : i + batch_size]
        ]
        db_conn.add_fetching_uids(uid_docs_batch)

    if len(db_conn.fetching_uids_batch) > 0:
        db_conn._push_fetching_uids_to_db()


def load_tids_from_file_to_db(files: list, batch_size: int, env_file: str, issue: str):
    """
    Load tids to database for hydration.

    Args:
        files (list): Files of tweet ids to load from.
        batch_size (int): Batch size for db updates.
        env_file (str): Dotenv file.
        issue (str): Name of issue for hydrating process.
    """
    settings.load_config(env_file=env_file)
    db_conn = db_client.DBClient(
        issue=issue,
    )

    for idx, filepath in enumerate(files):
        tids = _load_from_file(str(filepath))
        tid_docs = [{"tid": tid, "status": 0, "file": str(filepath)} for tid in tids]

        # insert all tids in batches of size batch_size
        for i in range(0, len(tid_docs), batch_size):
            db_conn.add_hydrating_tids(tid_docs[i : i + batch_size])

        if idx % 255 == 0:
            logger.info(f"Done with {idx+1}/{len(files)} files.")

    if len(db_conn.hydrating_tids_batch) > 0:
        db_conn._push_hydrating_tids_to_db()


def load_tids_to_db(
    tids: list,
    filepath: str = "unknown",
    db_conn: db_client.DBClient = None,
    issue: str = None,
    env_file: str = None,
    batch_size: int = 4096,
    nkeys: int = 9,
):
    keys = list(range(1, nkeys + 1))
    if not db_conn:
        settings.load_config(env_file=env_file)
        db_conn = db_client.DBClient(
            issue=issue,
        )

    for i in range(0, len(tids), batch_size):
        tid_docs_batch = [
            {"tid": tid, "status": 0, "file": str(filepath), "key": keys[i % nkeys]}
            for tid in tids[i : i + batch_size]
        ]
        db_conn.add_hydrating_tids(tid_docs_batch)

    if len(db_conn.hydrating_tids_batch) > 0:
        db_conn._push_hydrating_tids_to_db()


def _load_from_file(file):
    if file.endswith(".json"):
        with open(file, "r") as f:
            elements = json.load(f)
    elif file.endswith(".jsonl"):
        elements = []
        with open(file, "r") as f:
            for line in f:
                elements.append(json.loads(line))
    elif file.endswith(".pkl"):
        with open(file, "rb") as f:
            elements = pickle.load(f)
    elif file.endswith(".txt"):
        # read each line individually
        with open(file, "r") as f:
            elements = [line.strip() for line in f]
    else:
        raise errors.SnPipelineError(errors.SnPipelineErrorMsg.UNSUPPORTED_FILE_FORMAT)

    return elements


def _get_batches(elements, batch_size=4096):
    return [elements[i : i + batch_size] for i in range(0, len(elements), batch_size)]



import os
from multiprocessing import cpu_count
from pathlib import Path

from dotenv import load_dotenv

import tweepipe

MONGO_DB = None
MONGO_USERNAME = None
MONGO_PASSWORD = None
MONGO_HOST = None
MONGO_PORT = None

RESPONSE_MIRROR_COLLECTION_NAME = "raw_response"

CPU_COUNT = cpu_count() - 2
WORKER_COUNT = 8

STORE_TO_FILE = False

CHROME_DRIVER_EXE = None

SLACK_BOT_TOKEN = None
DATA_COLLECTION_CHANNEL_ID = None

CURRENT_TASK = None

ACADEMIC_API_BEARER_TOKEN = None
ACADEMIC_API_CONSUMER_KEY = None
ACADEMIC_API_CONSUMER_SECRET = None

SNPIPELINE_HOME = Path.cwd()
OUTPUT_DIR = SNPIPELINE_HOME
BOTSPOT_MODEL = Path("tweepipe/legacy/botspot/model/botspot_file.pkl")

TWEET_TS_STR_FORMAT = "%a %b %d %H:%M:%S +0000 %Y"
TWITTER_API_V2_TIME_STR = TWITTER_API_V1_STR_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
TWITTER_API_ACADEMIC_V2_STR_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def set_current_task(task: str = None):
    """
    Define the current task of the pipeline.

    Args:
        task (str, optional): Task name. Defaults to None.
    """
    tweepipe.settings.CURRENT_TASK = task


def load_config(env_file: str = None):
    """
    Set tweepipe global variables according to dotenv environment variables.

    Args:
        env_file (str, optional): Path to the dotenv file. Defaults to None.
    """
    load_dotenv(env_file, override=True)

    # environment variables to connect to mongodb
    tweepipe.settings.MONGO_DB = os.getenv("MONGO_DB")
    tweepipe.settings.MONGO_USERNAME = os.getenv("MONGO_USERNAME")
    tweepipe.settings.MONGO_PASSWORD = os.getenv("MONGO_PASSWORD")
    tweepipe.settings.MONGO_HOST = os.getenv("MONGO_HOST")
    tweepipe.settings.MONGO_PORT = int(os.getenv("MONGO_PORT", 27017))

    # connect to a slack channel to notify of data collection events
    tweepipe.settings.SLACK_BOT_TOKEN = os.getenv("SLACK_BOT_TOKEN")
    tweepipe.settings.DATA_COLLECTION_CHANNEL_ID = os.getenv("SLACK_CHANNEL_ID")

    # relative to the current home dir
    tweepipe.settings.OUTPUT_DIR = os.getenv("OUTPUT_DIR")
    if tweepipe.settings.OUTPUT_DIR != None:
        tweepipe.settings.OUTPUT_DIR = tweepipe.settings.SNPIPELINE_HOME.joinpath(
            tweepipe.settings.OUTPUT_DIR
        )
        tweepipe.settings.OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    else:
        tweepipe.settings.OUTPUT_DIR = tweepipe.settings.SNPIPELINE_HOME

    # environment variable to connect to academic api, use search tweets
    tweepipe.settings.ACADEMIC_API_BEARER_TOKEN = os.getenv("ACADEMIC_API_BEARER_TOKEN")
    tweepipe.settings.ACADEMIC_API_CONSUMER_KEY = os.getenv("ACADEMIC_API_CONSUMER_KEY")
    tweepipe.settings.ACADEMIC_API_CONSUMER_SECRET = os.getenv(
        "ACADEMIC_API_CONSUMER_SECRET"
    )

    # load standard
    tweepipe.settings.ACCESS_TOKEN = os.getenv("ACCESS_TOKEN")
    tweepipe.settings.ACCESS_TOKEN_SECRET = os.getenv("ACCESS_TOKEN_SECRET")
    tweepipe.settings.CONSUMER_KEY = os.getenv("CONSUMER_KEY")
    tweepipe.settings.CONSUMER_SECRET = os.getenv("CONSUMER_SECRET")
    tweepipe.settings.TWITTER_CREDENTIALS_V1 = {
        "access_token": tweepipe.settings.ACCESS_TOKEN,
        "access_token_secret": tweepipe.settings.ACCESS_TOKEN_SECRET,
        "consumer_key": tweepipe.settings.CONSUMER_KEY,
        "consumer_secret": tweepipe.settings.CONSUMER_SECRET,
    }

    # load path to chrome drive for debugging purposes
    tweepipe.settings.CHROME_DRIVER_EXE = os.getenv("CHROME_DRIVER_EXE")


def load_academic_api_credentials(env_file: str = None):
    """
    Directly load credentials to the academic api.

    Args:
        env_file (str, optional): Path to the dotenv file. Defaults to None.
    """
    load_dotenv(env_file, override=True)

    # environment variable to connect to academic api, use search tweets
    tweepipe.settings.ACADEMIC_API_BEARER_TOKEN = os.getenv("ACADEMIC_API_BEARER_TOKEN")
    tweepipe.settings.ACADEMIC_API_CONSUMER_KEY = os.getenv("ACADEMIC_API_CONSUMER_KEY")
    tweepipe.settings.ACADEMIC_API_CONSUMER_SECRET = os.getenv(
        "ACADEMIC_API_CONSUMER_SECRET"
    )



from pathlib import Path
import json

import tweepy
from tweepy import OAuthHandler
from loguru import logger

from tweepipe.db import db_client

TWITTER_API_PURPOSES = ["any", "lookup", "search", "stream", "hydrate"]


def fetch_issue_credentials(issue):
    filepath = Path(f".{issue}.json")
    if not filepath.is_file():
        raise FileNotFoundError(f"Cannot find credentials file for issue {issue}.")

    with open(filepath) as f:
        credentials = json.load(f)

    return credentials


def free_api(db_conn: db_client.DBClient, twitter_credentials: list, purpose: str):
    """Set the API status to 0 post collection."""
    for twitter_credential in twitter_credentials:
        db_conn.update_twitter_credentials(
            consumer_key=twitter_credential.get("consumer_key"),
            purpose=purpose,
        )


def _get_twitter_credentials(db_conn, api_count=-1, purpose="any", free_api=False):
    """Retrieve N twitter credentials from db."""

    if purpose not in TWITTER_API_PURPOSES:
        raise ValueError(
            "Invalid purpose. Choose among the following: {}.".format(
                TWITTER_API_PURPOSES
            )
        )

    return db_conn.get_twitter_credentials(
        count=api_count, purpose=purpose, must_be_free=free_api
    )


# TODO: check if instantaniation of twitter credentials valid
def _get_twitter_auth(credentials):
    twitter_auth = OAuthHandler(
        credentials["consumer_key"], credentials["consumer_secret"]
    )
    if "access_token" in credentials and "access_token_secret" in credentials:
        twitter_auth.set_access_token(
            credentials["access_token"], credentials["access_token_secret"]
        )

    return twitter_auth


def _get_twitter_auths(credentials):
    """Retrieve multiple twitter auths object."""

    twitter_auths = [_get_twitter_auth(cred_pair) for cred_pair in credentials]

    return twitter_auths


def _get_twitter_api(
    db_conn: db_client.DBClient = None, purpose: str = "", credentials: dict = {}
):
    """Get a single twitter api."""

    if not credentials and db_conn:
        credentials = _get_twitter_credentials(db_conn, api_count=1, purpose=purpose)[0]
    elif not credentials:
        raise RuntimeError("No credentials provided.")

    twitter_auth = _get_twitter_auth(credentials)
    twitter_api = tweepy.API(
        twitter_auth, wait_on_rate_limit=True, retry_count=10, retry_delay=3
    )

    return twitter_api


def _get_twitter_apis(
    db_conn: db_client.DBClient = None,
    purpose: str = "",
    api_count=-1,
    twitter_credentials: list = [],
    free_api: bool = False,
):
    """Get multiple twitter apis."""

    if not twitter_credentials and db_conn:
        twitter_credentials = _get_twitter_credentials(
            db_conn, api_count=api_count, purpose=purpose, free_api=free_api
        )
    elif not twitter_credentials:
        raise RuntimeError("No credentials provided.")

    logger.info(f"Retrieving {len(twitter_credentials)}")

    twitter_auths = _get_twitter_auths(twitter_credentials)

    twitter_apis = []
    for twitter_auth in twitter_auths:
        tmp_twitter_api = tweepy.API(
            twitter_auth, wait_on_rate_limit=True, wait_on_rate_limit_notify=True
        )
        twitter_apis.append(tmp_twitter_api)

    return twitter_apis, twitter_credentials



from loguru import logger
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from tweepipe import settings


class SlackLogger:
    """Log data collection events to slack."""

    @classmethod
    def log_start_task(cls, task: str, msg: str = None):
        """Log the begining of a new task to slack.

        :param task: Name of the started task.
        :type task: str
        :param msg: Message attached to starting the task, defaults to None.
        :type msg: str, optional
        """

        start_msg = f"Task: {task} - Message: {msg}"
        cls.post_message(start_msg)

    @classmethod
    def log_error(cls, error: str, msg: str = None):
        """Log error on ongoing data collection task.

        :param error: Error
        :type error: str
        :param msg: Error message, defaults to None.
        :type msg: str, optional
        """

        # build error msg depending on ongoing data collection
        error_msg = f"Task: {settings.CURRENT_TASK} - " if settings.CURRENT_TASK else ""
        error_msg += f"Error: {error} - Message: {msg}"
        cls.post_message(error_msg)

    @classmethod
    def post_message(cls, msg: str = None):
        # init client on each error (should be small)
        slack_client = WebClient(token=settings.SLACK_BOT_TOKEN)

        try:
            _ = slack_client.chat_postMessage(
                channel=settings.DATA_COLLECTION_CHANNEL_ID, text=msg
            )
        except SlackApiError as e:
            logger.error(f"Error occured while posting message: {e}.")



import datetime

import pymongo
from tweepipe.legacy.utils import extract
import tweepy
from loguru import logger

from tweepipe import settings
from tweepipe.db import db_client, db_schema
from tweepipe.utils import snowflake


def _run_hydration(
    twitter_api: tweepy.API,
    tweet_ids: list,
    issue: str,
    env_file: str = None,
    include_users: bool = True,
    include_relations: bool = True,
):
    """Execute hydration process."""

    if env_file:
        settings.load_config(env_file=env_file)

    db_conn = db_client.DBClient(
        issue=issue, include_relations=include_relations, include_users=include_users
    )

    hydrated_count = 0
    processed_count = 0
    missing_count = 0
    lookup_batch_size = 100
    for i in range(0, len(tweet_ids), lookup_batch_size):
        all_lookedup_tweets = twitter_api.statuses_lookup(
            tweet_ids[i : i + lookup_batch_size], tweet_mode="extended", map_=True
        )

        missing_tids = [
            tweet._json for tweet in all_lookedup_tweets if len(tweet._json) == 1
        ]
        lookedup_tweets = [
            tweet for tweet in all_lookedup_tweets if len(tweet._json) != 1
        ]

        for lookedup_tweet in lookedup_tweets:
            processed_tweet = extract.retrieve_content_from_tweet(
                lookedup_tweet._json,
                include_users=include_users,
                include_relations=include_relations,
                db_conn=db_conn,
            )

        # store all missing tweets, potentially from suspended accounts
        db_conn.add_missing_tids(missing_tids)

        missing_count += len(missing_tids)
        hydrated_count += len(lookedup_tweets)
        processed_count += len(all_lookedup_tweets)

    if db_conn:
        db_conn.flush_content()

    return (hydrated_count, missing_count, processed_count)


def handle_db_hydration(
    twitter_api: tweepy.API,
    issue: str,
    env_file: str = None,
    include_users: bool = True,
    include_relations: bool = True,
    target_db: str = None,
    collection: str = "hydrating_tids",
    batch_size: int = 1024,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    key: int = None,
):
    # hydrate all tweets from database, if error arises log it
    while True:
        try:
            _run_hydration_from_db(
                twitter_api=twitter_api,
                issue=issue,
                env_file=env_file,
                include_users=include_users,
                include_relations=include_relations,
                target_db=target_db,
                collection=collection,
                batch_size=batch_size,
                start_date=start_date,
                end_date=end_date,
                key=key,
            )
        except Exception as e:
            logger.error(f"Error arising from hydrating {e}.")


def _run_hydration_from_db(
    twitter_api: tweepy.API,
    issue: str,
    env_file: str = None,
    include_users: bool = True,
    include_relations: bool = True,
    target_db: str = None,
    collection: str = "hydrating_tids",
    batch_size: int = 1024,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    key: int = None,
):
    """Run hydration by taking tweets from database and hydrating them"""

    if env_file:
        settings.load_config(env_file=env_file)

    if not target_db:
        target_db = issue

    logger.info(f"Hydrating tweets from {collection} in {issue} to {target_db}.")

    # setup target database with collections
    db_conn = db_client.DBClient(
        issue=target_db,
        include_relations=include_relations,
        include_users=include_users,
        schema=db_schema.INDEX_V3,
    )
    hydrating_tids_collection = db_conn._get_collection(collection, db_name=issue)

    filter = {"status": 0}
    if start_date and end_date:
        start_tid = snowflake.get_tweet_id_from_time(start_date)
        end_tid = snowflake.get_tweet_id_from_time(end_date)
        tid_range_query = {"tid_int": {"$gte": start_tid, "$lt": end_tid}}
        filter.update(tid_range_query)
    if key:
        filter.update({"key": key})

    projection = {"tid": True, "_id": True}
    logger.info(f"Filtering tweets with the key {key} - full filter : {filter}.")
    cursor = hydrating_tids_collection.find(
        filter=filter, projection=projection, batch_size=batch_size, allow_disk_use=True
    )

    success_batch, missing_batch, lookup_batch = [], [], []
    for doc in cursor:
        hydrating_tids_collection.update_one(
            {"_id": doc["_id"]}, {"$set": {"status": 1}}
        )
        lookup_batch.append(doc)

        # max lookup batch size is 100
        if len(lookup_batch) == 100:
            batch_tid_id_map = {doc["tid"]: doc["_id"] for doc in lookup_batch}
            all_lookedup_tweets = twitter_api.statuses_lookup(
                [doc["tid"] for doc in lookup_batch], tweet_mode="extended", map_=True
            )

            for tweet in all_lookedup_tweets:
                if len(tweet._json) == 1:
                    missing_batch.append(
                        pymongo.operations.UpdateOne(
                            {"_id": batch_tid_id_map[str(tweet.id)]},
                            {"$set": {"status": -1}},
                        )
                    )
                else:
                    processed_tweet = extract.retrieve_content_from_tweet(
                        tweet._json,
                        include_users=include_users,
                        include_relations=include_relations,
                        db_conn=db_conn,
                    )
                    success_batch.append(
                        pymongo.operations.UpdateOne(
                            {"_id": batch_tid_id_map[str(tweet.id)]},
                            {"$set": {"status": 2}},
                        )
                    )

            lookup_batch = []

        # if enough, update status of tweets to 'error'
        if len(missing_batch) >= batch_size:
            try:
                result = hydrating_tids_collection.bulk_write(
                    missing_batch, ordered=False
                )
            except pymongo.errors.BulkWriteError as e:
                pass

            missing_batch = []

        # same, update status of tweets to 'done'
        if len(success_batch) >= batch_size:
            try:
                result = hydrating_tids_collection.bulk_write(
                    success_batch, ordered=False
                )
            except pymongo.errors.BulkWriteError as e:
                pass

            success_batch = []

    # if enough, update status of tweets to 'error'
    if len(missing_batch) > 0:
        try:
            result = hydrating_tids_collection.bulk_write(missing_batch, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass

        missing_batch = []

    # same, update status of tweets to 'done'
    if len(success_batch) > 0:
        try:
            result = hydrating_tids_collection.bulk_write(success_batch, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass

        success_batch = []


def _split_time_interval(start_date, end_date, n):
    dt = (end_date - start_date) / n
    dts = [(start_date + i * dt, start_date + (i + 1) * dt) for i in range(n)]

    return dts


def _get_parallel_hydrating_from_db_kwargs(
    twitter_apis: list,
    issue: str,
    target_db: str = None,
    env_file: str = None,
    include_relations: bool = True,
    include_users: bool = True,
    batch_size: int = 4096,
    collection: str = "hydrating_tids",
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
):
    """Get list of keyword args for the hydration process with tids
    sourced from the database."""

    if start_date and end_date:
        # split the time periods into len(twitter_apis) pieces
        dts = _split_time_interval(start_date, end_date, len(twitter_apis))
    else:
        keys = list(range(1, 1 + len(twitter_apis)))

    logger.info(f"Preparing workload distribution on keys {keys}.")

    kwargs_list = []
    iterator = (
        zip(twitter_apis, dts) if (start_date and end_date) else zip(twitter_apis, keys)
    )
    for twitter_api, dt in iterator:
        kwargs = {
            "twitter_api": twitter_api,
            "include_users": include_users,
            "include_relations": include_relations,
            "issue": issue,
            "target_db": target_db,
            "collection": collection,
            "env_file": env_file,
            "batch_size": batch_size,
        }
        if start_date and end_date:
            kwargs.update(
                {
                    "start_date": dt[0],
                    "end_date": dt[1],
                }
            )
        else:
            kwargs.update({"key": dt})

        kwargs_list.append(kwargs)

    return kwargs_list


def _get_parallel_hydrating_kwargs(
    tweet_ids: list,
    twitter_apis: list,
    issue: str,
    env_file: str = None,
    include_relations: bool = True,
    include_users: bool = True,
    batch_size: int = 4096,
):
    """Generate arg list used to hydrate provided tweet ids."""

    kwargs_list = []
    for idx, i in enumerate(range(0, len(tweet_ids), batch_size)):
        tweet_ids_batch = tweet_ids[i : i + batch_size]
        twitter_api = twitter_apis[idx % len(twitter_apis)]
        kwargs = {
            "twitter_api": twitter_api,
            "tweet_ids": tweet_ids_batch,
            "include_users": include_users,
            "include_relations": include_relations,
            "issue": issue,
            "env_file": env_file,
        }
        kwargs_list.append(kwargs)

    return kwargs_list
